@@ -18,9 +18,9 @@ problem = {
             "gamma": 0.999,
             "output": "/root/ray_results/debug/",
             "output_compress_columns": [],
-            "soft_horizon": False,
-            "no_done_at_end": False,
-            "horizon": 10000,
+            "batch_mode": "complete_episodes",
+            "horizon": 1000,
+            "metrics_smoothing_episodes": 5,
         },
         "stop": {"timesteps_total": 400000},
     },
@@ -36,7 +36,7 @@ rllib_setup = {
             "num_gpus": 1,
             "seed": ray.tune.randint(0, 10000000),
         },
-        "log_to_file": True,
+        # "log_to_file": True,
         "checkpoint_freq": 10000,
     },
 }
