@@ -1,30 +1,38 @@
 """Module to schedule experiments."""
 
-import ray.tune
-import ray.tune.integration.wandb
-import click
+import os
 import uuid
 import subprocess
 import pathlib
 import importlib.util
-import os
+import ray.tune
+import ray.tune.integration.wandb
+import ray.tune.integration.mlflow
+
 
 import beobench.experiment.definitions.utils
 import beobench.experiment.definitions.default
+import beobench.experiment.definitions.methods
+import beobench.experiment.definitions.envs
 import beobench.experiment.containers
 import beobench.utils
 
 
 def run(
     experiment_file: str = None,
+    agent_file: str = None,
+    method: str = None,
+    env: str = None,
     local_dir: str = "./beobench_results/ray_results",
     wandb_project: str = "",
     wandb_entity: str = "",
     wandb_api_key: str = "",
+    mlflow_name: str = None,
     use_gpu: bool = False,
     docker_shm_size: str = "2gb",
     no_additional_container: bool = False,
     use_no_cache: bool = False,
+    _dev_beobench_location: str = None,
 ) -> None:
     """Run experiment.
 
@@ -34,6 +42,12 @@ def run(
     Args:
         experiment_file (str, optional): File that defines experiment.
             Defaults to None.
+        agent_file (str, optional): File that defines custom agent. This script is
+            executed inside the gym container.
+        method (str, optional): RL method to use in experiment. This overwrites any
+            method that is set in experiment file. For example 'PPO'. Defaults to None.
+        env (str, optional): environment to apply method to in experiment. This
+            overwrites any env set in experiment file. Defaults to None.
         local_dir (str, optional): Directory to write experiment files to. This argument
             is equivalent to the `local_dir` argument in `tune.run()`. Defaults to
             `"./beobench_results/ray_results"`.
@@ -43,6 +57,7 @@ def run(
         wandb_api_key (str, optional): wandb API key. Defaults to None.
         use_gpu (bool, optional): whether to use GPU from the host system. Defaults to
             False.
+        mlflow_name (str, optional): name of MLflow experiment.
         docker_shm_size(str, optional): size of the shared memory available to the
             container. Defaults to '2gb'."
         no_additional_container (bool, optional): wether not to start another container
@@ -50,32 +65,22 @@ def run(
             is started to run experiments in.
         use_no_cache (bool, optional): whether to use cache to build experiment
             container.
+        _dev_beobench_location (str, optional): github path to beobench package. For
+            developement purpose only. This will install a custom beobench version
+            inside the experiment container. By default the latest PyPI version is
+            installed.
     """
 
-    # Load experiment definition file
-    if experiment_file is None:
-        experiment_def = beobench.experiment.definitions.default
-    else:
+    # Create a definition of experiment from inputs
+    if experiment_file is not None:
         experiment_file = pathlib.Path(experiment_file)
-        # import experiment definition file as module
-        spec = importlib.util.spec_from_file_location(
-            "experiment_definition",
-            str(experiment_file.absolute()),
-        )
-        experiment_def = importlib.util.module_from_spec(spec)
-        spec.loader.exec_module(experiment_def)
-
-        # check if all attributes present and replace with default if not
-        if not hasattr(experiment_def, "problem"):
-            experiment_def.problem = beobench.experiment.definitions.default.problem
-        if not hasattr(experiment_def, "method"):
-            experiment_def.method = beobench.experiment.definitions.default.method
-        if not hasattr(experiment_def, "rllib_setup"):
-            experiment_def.rllib_setup = (
-                beobench.experiment.definitions.default.rllib_setup
-            )
+    if agent_file is not None:
+        agent_file = pathlib.Path(agent_file)
+    experiment_def = _create_experiment_def(experiment_file, method, env)
 
     if no_additional_container:
+        # Execute experiment
+        # (this is usually reached from inside an experiment container)
 
         # Add wandb callback if sufficient information
         # TODO: add earlier check to see that also API key available
@@ -85,21 +90,30 @@ def run(
             raise ValueError(
                 "Only one of wandb_project or wandb_entity given, but both required."
             )
+        elif mlflow_name:
+            callbacks = [_create_mlflow_callback(mlflow_name)]
         else:
             callbacks = []
 
         # change RLlib setup if GPU used
-        rllib_setup = experiment_def.rllib_setup
         if use_gpu:
-            rllib_setup["rllib_experiment_config"]["config"]["num_gpus"] = 1
+            experiment_def["rllib_setup"]["rllib_experiment_config"]["config"][
+                "num_gpus"
+            ] = 1
 
         # run experiment in ray tune
-        run_in_tune(
-            problem_def=experiment_def.problem,
-            method_def=experiment_def.method,
-            rllib_setup=rllib_setup,
-            rllib_callbacks=callbacks,
-        )
+        if not agent_file:
+            run_in_tune(
+                problem_def=experiment_def["problem"],
+                method_def=experiment_def["method"],
+                rllib_setup=experiment_def["rllib_setup"],
+                rllib_callbacks=callbacks,
+            )
+        else:
+            # run custom RL agent
+            args = ["python -m", f"/tmp/beobench/{agent_file.name}"]
+            subprocess.check_call(args)
+
     else:
         # build and run experiments in docker container
 
@@ -110,7 +124,7 @@ def run(
 
         # docker setup
         image_tag = beobench.experiment.containers.build_experiment_container(
-            build_context=experiment_def.problem["problem_library"],
+            build_context=experiment_def["problem"]["problem_library"],
             use_no_cache=use_no_cache,
         )
 
@@ -127,8 +141,16 @@ def run(
                 f"{exp_file_abs}:{exp_file_on_docker}:ro",
             ]
 
+        if agent_file is not None:
+            ag_file_abs = agent_file.absolute()
+            ag_file_on_docker = f"/tmp/beobench/{agent_file.name}"
+            docker_flags += [
+                "-v",
+                f"{ag_file_abs}:{ag_file_on_docker}:ro",
+            ]
+
         # enable docker-from-docker access only for built-in boptest integration.
-        if experiment_def.problem["problem_library"] == "boptest":
+        if experiment_def["problem"]["problem_library"] == "boptest":
 
             # Create docker network (only useful if starting other containers)
             beobench.experiment.containers.create_docker_network("beobench-net")
@@ -159,6 +181,8 @@ def run(
             beobench_flags.append(f"--wandb-entity={wandb_entity}")
         if use_gpu:
             beobench_flags.append("--use-gpu")
+        if agent_file:
+            beobench_flags.append(f"--agent-file={agent_file}")
         beobench_flag_str = " ".join(beobench_flags)
 
         # if no wandb API key is given try to get it from env
@@ -166,6 +190,14 @@ def run(
             # this will return "" if env var not set
             wandb_api_key = os.getenv("WANDB_API_KEY", "")
 
+        cmd_list_in_container = [""]
+        # dev mode where custom beobench is installed directly from git
+        if _dev_beobench_location is not None:
+            cmd_list_in_container.append("pip uninstall --yes beobench")
+            cmd_list_in_container.append(f"pip install {_dev_beobench_location}")
+
+        cmd_in_container = " && ".join(cmd_list_in_container)
+
         args = [
             "docker",
             "run",
@@ -183,8 +215,8 @@ def run(
             "/bin/bash",
             "-c",
             (
-                f"export WANDB_API_KEY={wandb_api_key} && python -m "
-                f"beobench.experiment.scheduler {beobench_flag_str} "
+                f"export WANDB_API_KEY={wandb_api_key} {cmd_in_container} && "
+                f"beobench run {beobench_flag_str} "
                 "--no-additional-container && bash"
             ),
         ]
@@ -192,86 +224,6 @@ def run(
         subprocess.check_call(args)
 
 
-@click.command()
-@click.option(
-    "--experiment-file",
-    default=None,
-    help="File that defines beobench experiment.",
-    type=click.Path(exists=True, file_okay=True, dir_okay=False),
-)
-@click.option(
-    "--local-dir",
-    default="./beobench_results/ray_results",
-    help="Local directory to write results to.",
-    type=click.Path(exists=False, file_okay=False, dir_okay=True),
-)
-@click.option(
-    "--wandb-project",
-    default="",
-    help="Weights and biases project name to log runs to.",
-)
-@click.option(
-    "--wandb-entity",
-    default="",
-    help="Weights and biases entity name to log runs under.",
-)
-@click.option(
-    "--wandb-api-key",
-    default="",
-    help="Weights and biases API key.",
-)
-@click.option(
-    "--use-gpu",
-    is_flag=True,
-    help="Whether to use GPUs from the host system in experiment container.",
-)
-@click.option(
-    "--docker-shm-size",
-    default="2gb",
-    help="Size of shared memory available to experiment container.",
-)
-@click.option(
-    "--no-additional-container",
-    is_flag=True,
-    help="Do not run another container to do experiments in.",
-)
-@click.option(
-    "--use-no-cache",
-    is_flag=True,
-    help="Whether to use cache to build experiment container.",
-)
-def run_command(
-    experiment_file: str,
-    local_dir: str,
-    wandb_project: str,
-    wandb_entity: str,
-    wandb_api_key: str,
-    use_gpu: bool,
-    docker_shm_size: str,
-    no_additional_container: bool,
-    use_no_cache: bool,
-) -> None:
-    """Run experiment from command line.
-
-    Command line version of run function. This appears to be
-    the best (but not great) way to have a parallel python and command
-    line interface.
-
-    See https://stackoverflow.com/a/40094408.
-    """
-    run(
-        experiment_file=experiment_file,
-        local_dir=local_dir,
-        wandb_project=wandb_project,
-        wandb_entity=wandb_entity,
-        wandb_api_key=wandb_api_key,
-        use_gpu=use_gpu,
-        docker_shm_size=docker_shm_size,
-        no_additional_container=no_additional_container,
-        use_no_cache=use_no_cache,
-    )
-
-
 def run_in_tune(
     problem_def: dict, method_def: dict, rllib_setup: dict, rllib_callbacks: list = None
 ) -> ray.tune.ExperimentAnalysis:
@@ -309,7 +261,7 @@ def run_in_tune(
 
     # register the problem environment with ray tune
     # env_creator is a module available in experiment containers
-    import env_creator  # pylint: disable=import-outside-toplevel
+    import env_creator  # pylint: disable=import-outside-toplevel,import-error
 
     ray.tune.registry.register_env(
         problem_def["rllib_experiment_config"]["config"]["env"],
@@ -336,11 +288,114 @@ def _create_wandb_callback(
     wandb_project: str,
     wandb_entity: str,
 ):
+    """Create an RLlib weights and biases (wandb) callback.
+
+    Args:
+        wandb_project (str): name of wandb project.
+        wandb_entity (str): name of wandb entity that owns project.
+
+    Returns:
+        : a wandb callback
+    """
     wandb_callback = ray.tune.integration.wandb.WandbLoggerCallback(
         project=wandb_project, log_config=True, entity=wandb_entity
     )
     return wandb_callback
 
 
-if __name__ == "__main__":
-    run_command()  # pylint: disable=no-value-for-parameter
+def _create_mlflow_callback(
+    mlflow_name: str,
+):
+    """Create an RLlib MLflow callback.
+
+    Args:
+        mlflow_name (str, optional): name of MLflow experiment.
+
+    Returns:
+        : a wandb callback
+    """
+    mlflow_callback = ray.tune.integration.mlflow.MLflowLoggerCallback(
+        experiment_name=mlflow_name, tracking_uri="file:/root/ray_results/mlflow"
+    )
+    return mlflow_callback
+
+
+def _create_experiment_def(
+    experiment_file: pathlib.Path, method: str, env: str
+) -> dict:
+    """Create a Beobench experiment definition.
+
+    Args:
+        experiment_file (str): path to experiment file.
+        method (str): name of RL method.
+        env (str): name of environment.
+    """
+    experiment_def = _load_experiment_file(experiment_file)
+
+    # parsing high level interface options
+
+    # methods
+    if method:
+        if method == "ppo":
+            experiment_def["method"] = beobench.experiment.definitions.methods.PPO
+        else:
+            raise ValueError(
+                (
+                    f"The supplied method '{method}' does not match any of "
+                    "the pre-configured beobench methods."
+                )
+            )
+
+    if env:
+        if env == "boptest_bestest-hydronic-heat-pump-v1":
+            experiment_def["problem"] = getattr(
+                beobench.experiment.definitions.envs,
+                env.replace("-", "_"),
+            )
+        if env == "sinergym_eplus-5zone-hot-continous-v1":
+            experiment_def["problem"] = getattr(
+                beobench.experiment.definitions.envs,
+                env.replace("-", "_"),
+            )
+        if env == "energym_mixed-use-fan-fcu-v0":
+            experiment_def["problem"] = getattr(
+                beobench.experiment.definitions.envs,
+                env.replace("-", "_"),
+            )
+
+    return experiment_def
+
+
+def _load_experiment_file(experiment_file: pathlib.Path) -> dict:
+    """Load a Beobench experiment file.
+
+    Args:
+        experiment_file (str): path to experiment file.
+    """
+
+    # Load experiment definition file
+    if experiment_file is None:
+        experiment_file_mod = beobench.experiment.definitions.default
+    else:
+        # import experiment definition file as module
+        spec = importlib.util.spec_from_file_location(
+            "experiment_definition",
+            str(experiment_file.absolute()),
+        )
+        experiment_file_mod = importlib.util.module_from_spec(spec)
+        spec.loader.exec_module(experiment_file_mod)
+
+    experiment_def = dict()
+
+    # Create experiment def dictionary, and set default values if not available
+    # from experiment_file (module).
+    for exp_part in ["problem", "method", "rllib_setup"]:
+        if hasattr(experiment_file_mod, exp_part):
+            experiment_def[exp_part] = getattr(experiment_file_mod, exp_part)
+        else:
+            experiment_def[exp_part] = getattr(
+                beobench.experiment.definitions.default,
+                exp_part,
+            )
+
+    return experiment_def
